#include "korgcomp.hpp"
#include <string.h>
#include <ctype.h>
#include <stdlib.h>

static const char* MIDISECTION = "[MIDI]";
static const char* MIDIINITSECTION = "[MIDI-INIT]";
static const char* COMPATSECTION = "[COMPAT-";
static const char* COMPATOSCSECTION = "[COMPATOSC-";
static const char* NAMESECTION = "[KEYBOARD]";
static const char* BANKSECTION = "[BANK]";
static const char* DRUMBANK = "DRUM";
static const char* NAME = "NAME";
static char line[256];

MidiParam::MidiParam()
{
  hbank = lbankmin = lbankmax = progmin = progmax = 0;
}

MIDIParamDef::MIDIParamDef()
{
  bank = progmin = progmax = 0;
}

BankParam::BankParam()
{
  name[0] = 0;
  bank = progmin = progmax = 0;
  title[0] = 0;
}

ProgramParam::ProgramParam()
{
  bank = prog = 0;
  name[0] = 0;
}

CompatParam::CompatParam()
{
  bank = prog = 0;
  bankX = progX = 0;
}

CompatOscParam::CompatOscParam()
{
  osc = 0;
  oscX = 0;
}

KorgIProgramDef::KorgIProgramDef(const char* filename)
{
  keyboard_[0] = 0;
  f_ = fopen(filename, "r");
  if (f_)
  {
    if (findsection(NAMESECTION))
    {
      if (fgets(line, sizeof(line), f_) != 0 && line[0] != '[')
      {
        char* s = line;
        skipws(s);
        char* k = keyboard_;
        for (int i = 0; i < sizeof(keyboard_)-1 && isalnum(*s); i++)
          *k++ = *s++;
        *k = 0;
      }
    }
  }
}

FILE* KorgIProgramDef::getf()
{
  return f_;
}

KorgIProgramDef::~KorgIProgramDef()
{
  if (f_)
   fclose(f_);
  f_ = 0;
}

void KorgIProgramDef::skipws(char* &s)
{
  while (*s)
  {
    if (isspace(*s))
      s++;
    else if (*s == ';')
      s += strlen(s);
    else
      break;
  }
}

int KorgIProgramDef::matchchar(char*& s, int c)
{
  skipws(s);
  if (*s != c)
    return 0;
  s++;
  return 1;
}

int KorgIProgramDef::getnumber(char* &s)
{
  skipws(s);
  if (*s == 0)
    return -1;
  char* end;
  unsigned long v = strtoul(s, &end, 0);
  s = end;
  return (int)v;
}

int KorgIProgramDef::getnumberrange(char* &s, int& min, int& max)
{
  min = getnumber(s);
  if (min < 0)
    return 0;
  max = min;
  if (matchchar(s, '-'))
  {
    max = getnumber(s);
    if (max < 0)
      return 0;
  }
  return min <= max;
}

int KorgIProgramDef::getmidiparamdef(char* line, MIDIParamDef& param)
{
  char* s = line;

  param.bank = getnumber(s);
  if (param.bank < 0)
    return 0;
  if (!matchchar(s, ','))
    return 0;
  if (!getnumberrange(s, param.progmin, param.progmax))
    return 0;
  if (!matchchar(s, ':'))
    return 0;
  param.midi.hbank= getnumber(s);
  if (!matchchar(s, ','))
    return 0;
  if (!getnumberrange(s, param.midi.lbankmin, param.midi.lbankmax))
    return 0;
  if (!matchchar(s, ','))
    return 0;
  if (!getnumberrange(s, param.midi.progmin, param.midi.progmax))
    return 0;
  if (param.progmax - param.progmin > 1)
  if (param.progmax - param.progmin != param.midi.progmax - param.midi.progmin)
    return 0;
  return 1;
}

int KorgIProgramDef::getMIDIParam(int bank, int prog, MidiParam& param)
{
  if (!f_)
    return 0;
  if (!findsection(MIDISECTION))
    return 0;
  while (fgets(line, sizeof(line), f_) != 0 && line[0] != '[')
  {
    MIDIParamDef paramdef;
    if (!getmidiparamdef(line, paramdef))
      continue;
    if (bank == paramdef.bank &&
       prog >= paramdef.progmin &&
       prog <= paramdef.progmax)
    {
      param = paramdef.midi;
      param.progmin = param.progmax = prog - paramdef.progmin + paramdef.midi.progmin;
      return 1;
    }
  }
  return 0;
}

int KorgIProgramDef::getKeyboardParam(MidiParam& param, int &bank, int &prog, int needdrum)
{
  if (!f_)
    return 0;
  if (!findsection(MIDISECTION))
    return 0;
  while (fgets(line, sizeof(line), f_) != 0 && line[0] != '[')
  {
    MIDIParamDef paramdef;
    if (!getmidiparamdef(line, paramdef))
      continue;
    if (param.hbank == paramdef.midi.hbank &&
       param.lbankmin >= paramdef.midi.lbankmin &&
       param.lbankmin <= paramdef.midi.lbankmax &&
       param.progmin >= paramdef.midi.progmin &&
       param.progmax <= paramdef.midi.progmax &&
       (paramdef.progmax == paramdef.progmin ||
       paramdef.progmax - paramdef.progmin == paramdef.midi.progmax - paramdef.midi.progmin))
    {
      bank = paramdef.bank;
      prog = param.progmin - paramdef.midi.progmin + paramdef.progmin;

      if (needdrum)
      {
        if (getDrumBank() != bank)
        {
          bank = getDrumBank();
          prog = 0;
          continue;
        }
      }
      return 1;
    }
  }
  return 0;
}

int KorgIProgramDef::getDrumBank()
{
  BankParam bankinfo;
  if (getBankParam(DRUMBANK, bankinfo))
    return bankinfo.bank;
  return -1; // unknown
}


int KorgIProgramDef::getDrumBank(BankParam& bankinfo)
{
  return getBankParam(DRUMBANK, bankinfo);
}

int KorgIProgramDef::isUserProg(int bank, int prog)
{
  BankParam info;
  if (!getBankParam(bank, prog, info))
    return 0;
  if (strncmp(info.name, "USER", 4) == 0)
    return 1;
  return 0;
}

int KorgIProgramDef::getBankParam(const char *name, BankParam& bankinfo)
{
  if (!f_ || !name)
    return 0;
  if (!findsection(BANKSECTION))
    return 0;
  while (fgets(line, sizeof(line), f_) != 0)
  {
    char* s = line;
    if (!getbankdef(s, bankinfo))
      continue;
    if (strcmp(bankinfo.name, name) == 0)
      return 1;
  }
  return 0;
}

int KorgIProgramDef::getprogramdef(char*& s, ProgramParam& proginfo)
{
  skipws(s);
  int i;
  proginfo.bank = getnumber(s);
  if (proginfo.bank < 0)
    return 0;
  if (!matchchar(s, ','))
    return 0;
  proginfo.prog = getnumber(s);
  if (proginfo.prog < 0)
    return 0;
  if (!matchchar(s, '='))
    return 0;
  char* t = proginfo.name;
  skipws(s);
  for (i = 0; *s && *s != '\r' && *s != '\n' && i < sizeof(proginfo.name)-1; i++)
    *t++ = *s++;
  *t = 0;
  return 1;
}

int KorgIProgramDef::getbankdef(char* &s, BankParam& bankinfo)
{
  skipws(s);
  int i;
  char* t = bankinfo.name;
  for (i = 0; *s && isalnum(*s) && i < sizeof(bankinfo.name)-1; i++)
    *t++ = *s++;
  *t = 0;

  if (!matchchar(s, '='))
    return 0;
  bankinfo.bank = getnumber(s);
  if (bankinfo.bank < 0)
    return 0;
  if (!matchchar(s, ','))
    return 0;
  if (!getnumberrange(s, bankinfo.progmin, bankinfo.progmax))
    return 0;
  if (!matchchar(s, ','))
    return 0;
  t = bankinfo.title;
  skipws(s);
  for (i = 0; *s && isalnum(*s) && i < sizeof(bankinfo.title)-1; i++)
    *t++ = *s++;
  *t = 0;
  return 1;
}

int KorgIProgramDef::getBankParam(int bank, int prog, BankParam& bankinfo)
{
  if (!f_)
    return 0;
  if (!findsection(BANKSECTION))
    return 0;
  while (fgets(line, sizeof(line), f_) != 0)
  {
    char* s = line;
    if (!getbankdef(s, bankinfo))
      continue;
    if (bank == bankinfo.bank &&
        prog >= bankinfo.progmin &&
        prog <= bankinfo.progmax)
      return 1;
  }
  return 0;
}

int KorgIProgramDef::findsection(const char* s)
{
  if (!f_)
    return 0;
  rewind(f_);
  while (fgets(line, sizeof(line), f_) != 0)
  {
    if (strstr(line, s))
      return 1;
  }
  return 0;
}

int KorgIProgramDef::getCompatiblei3(int bank, int prog, int& banki3, int& progi3)
{
  return getCompatible("i3", bank, prog, banki3, progi3);
}

int KorgIProgramDef::getcompatdef(char* &s, CompatParam& compat)
{
  compat.bank = getnumber(s);
  if (compat.bank < 0)
    return 0;
  if (!matchchar(s, ','))
    return 0;
  compat.prog = getnumber(s);
  if (compat.prog < 0)
    return 0;
  if (!matchchar(s, ':'))
    return 0;
  compat.bankX = getnumber(s);
  if (compat.bankX < 0)
    return 0;
  if (!matchchar(s, ','))
    return 0;
  compat.progX = getnumber(s);
  if (compat.progX < 0)
    return 0;
  return 1;
}

int KorgIProgramDef::getCompatible(const char* keyboard, int bank, int prog, int& bankX, int& progX)
{
  bankX = 0;
  progX = prog;

  if (!f_)
    return 0;
  char compat[20];
  strcpy(compat,COMPATSECTION);
  strcat(compat, keyboard);
  strcat(compat, "]");
  if (!findsection(compat))
    return 0;
  while (fgets(line, sizeof(line), f_) != 0 && line[0] != '[')
  {
    char* s = line;
    CompatParam param;
    if (!getcompatdef(s, param))
      continue;
    if (bank == param.bank && prog == param.prog)
    {
      bankX = param.bankX;
      progX = param.progX;
      return 1;
    }
  }
  return 0;
}

int KorgIProgramDef::getReverseCompatible(const char* keyboard, int bankX, int progX, int& bank, int& prog)
{
  bank = 0;
  prog = prog;

  if (!f_)
    return 0;
  char compat[20];
  strcpy(compat,COMPATSECTION);
  strcat(compat, keyboard);
  strcat(compat, "]");
  if (!findsection(compat))
    return 0;
  while (fgets(line, sizeof(line), f_) != 0 && line[0] != '[')
  {
    char* s = line;
    CompatParam param;
    if (!getcompatdef(s, param))
      continue;
    if (bankX == param.bankX && progX == param.progX)
    {
      bank = param.bank;
      prog = param.prog;
      return 1;
    }
  }
  return 0;
}

const char* KorgIProgramDef::getName()
{
  return keyboard_;
}

const char* KorgIProgramDef::getProgramName(int bank, int prog)
{
  if (!findsection(NAME))
    return 0;
  while (fgets(line, sizeof(line), f_) != 0 && line[0] != '[')
  {
    char* s = line;
    ProgramParam param;
    if (!getprogramdef(s, param))
      continue;
    if (param.bank == bank && param.prog == prog)
    {
      strcpy(line, param.name);
      return line;
    }
  }
  return 0;
}

int KorgIProgramDef::findProgram(const char* name, int& bank, int& prog)
{
  if (!name || !*name)
    return 0;
  if (!findsection(NAME))
    return 0;
  while (fgets(line, sizeof(line), f_) != 0 && line[0] != '[')
  {
    char* s = line;
    ProgramParam param;
    if (!getprogramdef(s, param))
      continue;
    if (strcmp(name, param.name) == 0)
    {
      bank = param.bank;
      prog = param.prog;
      return 1;
    }
  }
  return 0;
}

int KorgIProgramDef::getMIDIInit(unsigned char* data, int maxlen)
{
  if (!data)
    return 0;
  if (!findsection(MIDIINITSECTION))
    return 0;
  int len = 0;
  unsigned char* nextdata = data;
  unsigned char* d;
  while (fgets(line, sizeof(line), f_) != 0 && line[0] != '[')
  {
    if (len >= maxlen)
      break;
    *nextdata = 0; // length byte
    d = nextdata+1;

    int n = 0;
    char* s = line;
    while (*s)
    {
      int v = getnumber(s);
      if (v < 0 || v >= 256)
      {
         n = 0;
         break;
      }
      if (len + 1 + n >= maxlen)
      {
        n = 0;  // too large
        break;
      }
      *d++ = v;
      n++;
      skipws(s);
    }
    if (n > 0)
    {
      *nextdata = n;
      nextdata += n+1;
      len += n+1;
    }
  }
  return len;
}

int KorgIProgramDef::getcompatoscdef(char* &s, CompatOscParam& compatosc)
{
  compatosc.osc = getnumber(s);
  if (compatosc.osc < 0)
    return 0;
  if (!matchchar(s, ':'))
    return 0;
  compatosc.oscX = getnumber(s);
  if (compatosc.oscX < 0)
    return 0;
  return 1;
}


int KorgIProgramDef::getCompatiblei3Oscillator(int osc, int& osci3)
{
  return getCompatibleOscillator("i3", osc, osci3);
}

int KorgIProgramDef::getCompatibleOscillator(const char* keyboard, int osc, int& oscX)
{
  oscX = 0;

  if (!f_)
    return 0;
  char compat[20];
  strcpy(compat,COMPATOSCSECTION);
  strcat(compat, keyboard);
  strcat(compat, "]");
  if (!findsection(compat))
    return 0;
  while (fgets(line, sizeof(line), f_) != 0 && line[0] != '[')
  {
    char* s = line;
    CompatOscParam param;
    if (!getcompatoscdef(s, param))
      continue;
    if (osc == param.osc)
    {
      oscX = param.oscX;
      return 1;
    }
  }
  return 0;
}

int KorgIProgramDef::getReverseCompatibleOscillator(const char* keyboard, int oscX, int& osc)
{
  osc = 0;

  if (!f_)
    return 0;
  char compat[20];
  strcpy(compat,COMPATOSCSECTION);
  strcat(compat, keyboard);
  strcat(compat, "]");
  if (!findsection(compat))
    return 0;
  while (fgets(line, sizeof(line), f_) != 0 && line[0] != '[')
  {
    char* s = line;
    CompatOscParam param;
    if (!getcompatoscdef(s, param))
      continue;
    if (oscX == param.oscX)
    {
      osc = param.osc;
      return 1;
    }
  }
  return 0;
}

int getCompatible(KorgIProgramDef& def, int bank, int prog, KorgIProgramDef& defx, int &bankx, int &progx)
{
  if (def.getCompatible(defx.getName(), bank, prog, bankx, progx))
    return 1;

  // try to find identical bank
  BankParam param, paramx;
  if (def.getBankParam(bank, prog, param))
  if (defx.getBankParam(param.name, paramx))
  {
    if (param.progmax - param.progmin == paramx.progmax - paramx.progmin)
    {
      // bank compatible!
      bankx = paramx.bank;
      progx = prog - param.progmin + paramx.progmin;
      return 1;
    }
  }

  if (defx.getReverseCompatible(def.getName(), bank, prog, bankx, progx))
  {
    return 1;
  }
  if (def.getDrumBank(param))
  {
    if (bank == param.bank && defx.getDrumBank(paramx))
    {
      bankx = paramx.bank;
      progx = paramx.progmin;
      return 1;
    }
  }


  bankx = 0;
  progx = prog % 64;
  return 0;
}

int getCompatibleOscillator(KorgIProgramDef& def, int osc, KorgIProgramDef& defx, int& oscx)
{
  if (def.getCompatibleOscillator(defx.getName(), osc, oscx))
    return 1;

  if (defx.getReverseCompatibleOscillator(def.getName(), osc, oscx))
  {
    return 1;
  }
  osc = oscx;
  return 0;
}